import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from woffl.pvt.blackoil import BlackOil

# only works if the command python -m tests.boil_test is used
dirname = os.getcwd()
filename = os.path.join(dirname, r"data\oil_22_api_hysys_peng_rob.xlsx")  # , r"data\methane_hysys_peng_rob.xlsx")

hys_df = pd.read_excel(filename, header=1)

prs_ray = hys_df["pressure"]
temp = 80
hy_rho = hys_df["density"]
# hy_z_fact = hys_df["z_factor"]
hy_visc = hys_df["viscosity"]

py_rho = []
py_solub = []
py_visc = []
py_fvf = []
py_comp = []

py_boil = BlackOil.test_oil()

for prs in prs_ray:
    py_boil = py_boil.condition(prs, temp)
    py_rho.append(py_boil.density)
    py_solub.append(py_boil.gas_solubility())
    py_visc.append(py_boil.viscosity())
    py_fvf.append(py_boil.oil_fvf())
    py_comp.append(py_boil.compress())

fig, axs = plt.subplots(5, sharex=True)

axs[0].scatter(prs_ray, hy_rho, label="hysys")
axs[0].scatter(prs_ray, py_rho, label="python")
axs[0].set_ylabel("Density, lbm/ft3")
axs[0].legend()

axs[1].scatter(prs_ray, hy_visc, label="hysys")
axs[1].scatter(prs_ray, py_visc, label="python")
axs[1].set_ylabel("Viscosity, cP")
axs[1].legend()

# axs[2].scatter(prs_ray, hy_z_fact, label="hysys")
axs[2].scatter(prs_ray, py_solub, label="python")
axs[2].set_ylabel("Gas Solubility, SCF/STB")
axs[2].legend()
# axs[2].set_xlabel("Pressure, psig")

axs[3].scatter(prs_ray, py_fvf, label="python")
axs[3].set_ylabel("Oil Formation Vol. Factor, RB/STB")

ycoord = (max(py_comp) - min(py_comp)) / 10
axs[4].scatter(prs_ray, py_comp, label="python")
axs[4].set_ylabel("Oil Compressibility, psi^-1")
axs[4].axvline(x=py_boil.pbp, color="black", linestyle="--", linewidth=1)
axs[4].annotate(text="Bubble Point", xy=(py_boil.pbp, ycoord), rotation=90)
axs[4].set_xlabel("Pressure, psig")

fig.suptitle(f"Oil {py_boil.oil_api} API Properties at {temp} deg F")
plt.show()

print(f"Oil Surface Tension: {round(py_boil.tension() / 0.0000685, 2)} dyne/cm")



"""Batch Jet Pump Runs

Contains code that is used to run multiple pumps at once to understand the
current conditions. Currently Scott's code is here with a simple nested for
loop. Will eventually update this to more of a class style system. Runs the
analysis and sends the results to a .csv file.
"""

from dataclasses import dataclass
from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import woffl.assembly.sysops as so
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.resmix import ResMix


@dataclass(frozen=True)
class BatchResult:
    """Code is currently not used"""

    wellname: str
    nozzle: str
    throat: str
    choked: bool
    qoil_std: float
    fwat_bwpd: float
    mach_te: float
    total_water: float
    total_wc: float


class BatchPump:
    "Used for running different nozzle and throat combinations and selecting the correct one."

    def __init__(
        self,
        pwh: float,
        tsu: float,
        rho_pf: float,
        ppf_surf: float,
        wellbore: Pipe,
        wellprof: WellProfile,
        ipr_su: InFlow,
        prop_su: ResMix,
        wellname: str = "na",
    ) -> None:
        """Batch Pump Solver

        Used for iterating across a wide range of different pumps. An adjacent dataclass will
        be made for storing the results. Could add an optional wellname here if desired?

        Args:
            pwh (float): Pressure Wellhead, psig
            tsu (float): Temperature Suction, deg F
            rho_pf (float): Density of the power fluid, lbm/ft3
            ppf_surf (float): Pressure Power Fluid Surface, psig
            wellbore (Pipe): Pipe Class of the Wellbore
            wellprof (WellProfile): Well Profile Class
            ipr_su (InFlow): Inflow Performance Class
            prop_su (ResMix): Reservoir Mixture Conditions
        """
        self.pwh = pwh
        self.tsu = tsu
        self.rho_pf = rho_pf
        self.ppf_surf = ppf_surf
        self.wellbore = wellbore
        self.wellprof = wellprof
        self.ipr_su = ipr_su
        self.prop_su = prop_su
        self.wellname = wellname

    def update_pressure_wellhead(self, pwh: float) -> None:
        """Update the Wellhead Pressure

        Used to update the wellhead pressure instead of re-initializing everything.

        Args:
            pwh (float): Pressure Wellhead, psig
        """
        self.pwh = pwh

    def update_pressure_powerfluid(self, ppf_surf: float) -> None:
        """Update the Powerfluid Surface Pressure

        Used to update the powerfluid surface pressure instead of re-initializing everything.

        Args:
            ppf_surf (float): Pressure Power Fluid at Surface, psig
        """
        self.ppf_surf = ppf_surf

    def update_pressure_reservoir(self, pres: float) -> None:
        """Update the Reservoir Pressure in IPR

        Used to update reservoir pressure instead of re-initalizing everything.

        Args:
            pres (float): Pressure Reservoir, psig
        """
        self.ipr_su.pres = pres

    @staticmethod
    def jetpump_list(
        nozzles: list[str],
        throats: list[str],
        knz: float = 0.01,
        ken: float = 0.03,
        kth: float = 0.3,
        kdi: float = 0.4,
    ) -> list[JetPump]:
        """Create a list of Jet Pumps

        Automatically generate a list of different types of jet pumps. The list is fed into
        the batch run to understand performance. Could move this into batch run if desired.

        Args:
            nozzles (list): List of Nozzle Strings
            throats (list): List of Throat Ratios
            knz (float): Nozzle Friction Factor
            ken (float): Enterance Friction Factor
            kth (float): Throat Friction Factor
            kdi (float): Diffuser Friction Factor

        Returns:
            jp_list (list): List of JetPumps
        """
        jp_list = []
        for nozzle, throat in product(nozzles, throats):
            jp_list.append(JetPump(nozzle, throat, knz, ken, kth, kdi))
        return jp_list

    def batch_run(
        self,
        jetpumps: list[JetPump],
    ) -> list:
        """Batch Run of Jet Pumps

        Run through multiple different types of jet pumps. Results will be stored in
        a data class where the results can be graphed and selected for the optimal pump.

        Args:
            jetpump_list (list): List of JetPumps

        Returns:
            results (list): List of Dictionaries of Jet Pump Results
        """
        results = []
        for jetpump in jetpumps:
            try:
                psu_solv, sonic_status, qoil_std, fwat_bwpd, qnz_bwpd, mach_te = so.jetpump_solver(
                    self.pwh,
                    self.tsu,
                    self.rho_pf,
                    self.ppf_surf,
                    jetpump,
                    self.wellbore,
                    self.wellprof,
                    self.ipr_su,
                    self.prop_su,
                )
                result = {
                    "wellname": self.wellname,
                    "res_pres": self.ipr_su.pres,
                    "pf_pres": self.ppf_surf,
                    "nozzle": jetpump.noz_no,
                    "throat": jetpump.rat_ar,
                    "psu_solv": psu_solv,
                    "sonic_status": sonic_status,
                    "qoil_std": qoil_std,
                    "fwat_bwpd": fwat_bwpd,
                    "qnz_bwpd": qnz_bwpd,
                    "mach_te": mach_te,
                    "total_water": fwat_bwpd + qnz_bwpd,
                    "total_wc": fwat_bwpd + qnz_bwpd / (fwat_bwpd + qnz_bwpd + qoil_std),
                    "error": "na",
                }
            except Exception as exc:
                result = {
                    "wellname": self.wellname,
                    "res_pres": self.ipr_su.pres,
                    "pf_pres": self.ppf_surf,
                    "nozzle": jetpump.noz_no,
                    "throat": jetpump.rat_ar,
                    "psu_solv": np.nan,
                    "sonic_status": np.nan,
                    "qoil_std": np.nan,
                    "fwat_bwpd": np.nan,
                    "qnz_bwpd": np.nan,
                    "mach_te": np.nan,
                    "total_water": np.nan,
                    "total_wc": np.nan,
                    "error": exc,
                }
            results.append(result)  # add some progress bar code here?
        return results


# create a couple small functions that could be used across a pandas dataframe later
# need graphing, cleaning, and dropping variables, calculating gradients, finalized picking?


def batch_results_mask(
    qoil_std: list[float] | np.ndarray | pd.Series,
    qwat_tot: list[float] | np.ndarray | pd.Series,
) -> list[bool]:
    """Batch Results Mask

    Create a mask of booleans for the batch results that can be passed into either a dataframe,
    numpy array or list to filter out the unnecessary data. The unnecessary data are points where
    either a np.nan exists, or the oil rate is lower for a higher amount of water. The filtered
    points can then be passed to another function to calculate the gradient.

    Args:
        qoil_std (list): Oil Prod. Rate, BOPD
        qwat_tot (list): Total Water Rate, BWPD

    Returns:
        mask (list): True is a point to calc gradient, False means a point exists with better oil and less water
    """
    mask = []
    oil_copy = qoil_std.copy()
    wat_copy = qwat_tot.copy()

    for oil_main, wat_main in zip(qoil_std, qwat_tot):
        if oil_main == np.nan:
            status = False
        else:
            for oil_comp, wat_comp in zip(oil_copy, wat_copy):  # oil and water compare
                oil_diff = oil_comp - oil_main
                wat_diff = wat_comp - wat_main
                if (oil_diff > 0) and (wat_diff < 0):  # if another point has more oil for less water
                    status = False  # trash that point and exit the first loop
                    break
                else:
                    status = True  # has to maintain the True status for the entire loop through
        mask.append(status)
    return mask


def batch_results_plot(
    qoil_std: list[float] | np.ndarray | pd.Series,
    qwat_tot: list[float] | np.ndarray | pd.Series,
    nozzles: list[str] | np.ndarray | pd.Series,
    throats: list[str] | np.ndarray | pd.Series,
    wellname: str = "na",
    mask: list[bool] = [],
) -> None:
    """Batch Results Plot

    Create a plot to view the results from the batch run.

    Args:
        qoil_std (list): Oil Prod. Rate, BOPD
        qwat_tot (list): Total Water Rate, BWPD
        nozzles (list): Nozzle Numbers
        throats (list): Throat Ratios
        wellname (str): Wellname String
        mask (list): List of Booleans used for filtering data
    """
    jp_names = [noz + thr for noz, thr in zip(nozzles, throats)]  # create a list of all the jetpump names
    fig = plt.figure()
    ax = fig.add_subplot(111)

    if not mask:  # if the mask is an empty list
        mask = [False for noz in nozzles]  # create a list of falses

    for oil, water, jp, status in zip(qoil_std, qwat_tot, jp_names, mask):
        if oil == np.nan:
            pass
        else:
            if status:  # booleaan true of false
                ax.plot(water, oil, marker="o", linestyle="", color="r")  # one of the main point
            else:
                ax.plot(water, oil, marker="o", linestyle="", color="b")  # a one optimized point

            ax.annotate(jp, xy=(water, oil), xycoords="data", xytext=(1.5, 1.5), textcoords="offset points")

    ax.set_xlabel("Total Water Rate, BWPD")
    ax.set_ylabel("Produced Oil Rate, BOPD")

    if wellname == "na":
        ax.title.set_text("Jet Pump Performance")
    else:
        ax.title.set_text(f"{wellname} Jet Pump Performance")

    plt.show()



import math

import numpy as np

# eventually make dataclass called PumpCatalog, with different manufacturers


class JetPump:
    # Champion X jet pump nozzle sizes, inches
    nozzle_dia = [
        0.0553,
        0.0628,
        0.0705,
        0.0798,
        0.0903,  # matches 1 throat
        0.1016,  # matches 2 throat
        0.1145,  # doesn't match the 3 throat
        0.1291,  # matches 4 throat
        0.1458,  # matches 5 throat
        0.1643,
        0.1858,
        0.2099,
        0.237,
        0.2675,
        0.3017,
        0.3404,
        0.3841,
        0.4335,
        0.4981,
        0.5519,
    ]

    # Champion X jet pump throat sizes, inches
    throat_dia = [
        0.0903,
        0.1016,
        0.1151,
        0.1291,
        0.1458,
        0.1643,
        0.1858,
        0.2099,
        0.237,
        0.2675,
        0.3017,
        0.3404,
        0.3841,
        0.4335,
        0.4981,
        0.5519,
        0.6228,
        0.7027,
        0.7929,
        0.8947,
    ]

    area_code = {"X": -1, "A": 0, "B": 1, "C": 2, "D": 3, "E": 4}

    def __init__(
        self,
        nozzle_no: str,
        area_ratio: str,
        knz: float = 0.01,
        ken: float = 0.03,
        kth: float = 0.3,
        kdi: float = 0.3,
    ) -> None:
        """Jet Pump Initialization

        Create a Jet Pump object.

        Args:
            nozzle_no (string): Represents the Nozzle Size
            area_ratio (string): Represents the Area Ratio
            knz (float): Nozzle Friction Factor, unitless
            ken (float): Enterance Friction Factor, unitless
            kth (float): Throat Friction Factor, unitless
            kdi (float): Diffuser Friction Factor, unitless

        Returns:
            None
        """

        self.noz_no = str(nozzle_no)
        self.rat_ar = str(area_ratio)

        self.knz = knz
        self.ken = ken
        self.kth = kth
        self.kdi = kdi

        nozzle_idx = int(nozzle_no) - 1
        throat_idx = nozzle_idx + JetPump.area_code[area_ratio]

        dnz = None
        dth = None

        try:
            dnz = JetPump.nozzle_dia[nozzle_idx]
        except IndexError:
            print(f"Nozzle Size {nozzle_no} not recognized")

        try:
            dth = JetPump.throat_dia[throat_idx]
        except IndexError:
            print(f"Nozzle Throat Combo {str(nozzle_no)+str(area_ratio)} not recognized")

        self.dnz = dnz
        self.dth = dth

    def __repr__(self):
        return f"{self.noz_no+self.rat_ar} Jet Pump, Nozzle: {self.dnz} inches, Throat: {self.dth} inches"

    @staticmethod
    def area_circle(diameter):
        """Area of a Circle

        Args:
            diameter (float): diameter of a circle

        Returns:
            area (float): area of a circle
        """
        area = math.pi * (diameter**2) / 4
        return area

    @property
    def anz(self) -> float:
        """Area of Nozzle, ft2"""
        return self.area_circle(self.dnz) / 144

    @property
    def ath(self) -> float:
        """Area of Throat, ft2"""
        return self.area_circle(self.dth) / 144

    @property
    def ate(self) -> float:
        """Throat Enterance Area, ft2"""
        return self.ath - self.anz



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# from flow import jetflow as jf # for actual flow conditions
from woffl.pvt import BlackOil, FormGas, FormWater, ResMix

# test python reservoir mixture equations vs hysys
# only works if the command python -m tests.rmix_test is used
# q_oil_std = 100  # bopd

temp = 80
oil_api = 22
pbub = 1750  # psig
gas_sg = 0.55  # methane
wc = 0.3  # watercut
fgor = 800  # scf/stb

# mass fraction, volm fraction, mixture density, speed of sound (no hysys...)
dirname = os.getcwd()
filename = os.path.join(dirname, r"data\resmix_hysys_peng_rob.xlsx")

hy_df = pd.read_excel(filename, header=1)

prs_ray = hy_df["press"]
hy_mfac = [hy_df["oil_mfac"], hy_df["wat_mfac"], hy_df["gas_mfac"]]
hy_vfac = [hy_df["oil_vfac"], hy_df["wat_vfac"], hy_df["gas_vfac"]]
hy_rho_mix = hy_df["rho_mix"]

py_oil = BlackOil(oil_api, pbub, gas_sg)
py_wat = FormWater(wat_sg=1)
py_gas = FormGas(gas_sg)
py_mix = ResMix(wc, fgor, py_oil, py_wat, py_gas)

plot_names = ["Oil", "Wat", "Gas"]

py_mfac = []
py_vfac = []
py_rho_mix = []
py_cmix = []

for prs in prs_ray:
    py_mix = py_mix.condition(prs, temp)
    py_mfac.append(py_mix.mass_fract())
    py_vfac.append(py_mix.volm_fract())
    py_rho_mix.append(py_mix.rho_mix())
    py_cmix.append(py_mix.cmix())

# change formating to make it more user friendly to print
py_mfac = list(zip(*py_mfac))
py_vfac = list(zip(*py_vfac))

fig, axs = plt.subplots(4, sharex=True)

for i, hy in enumerate(hy_mfac):
    axs[0].scatter(prs_ray, hy, label=f"Hysys {plot_names[i]}")
    axs[0].scatter(prs_ray, py_mfac[i], marker="*", label=f"Python {plot_names[i]}")
axs[0].set_ylabel("Mass Fraction")
axs[0].legend()

for i, hy in enumerate(hy_vfac):
    axs[1].scatter(prs_ray, hy, label=f"Hysys {plot_names[i]}")
    axs[1].scatter(prs_ray, py_vfac[i], marker="*", label=f"Python {plot_names[i]}")
axs[1].set_ylabel("Volume Fraction")
axs[1].legend()

axs[2].scatter(prs_ray, hy_rho_mix, label="hysys")
axs[2].scatter(prs_ray, py_rho_mix, label="python")
axs[2].set_ylabel("Mixture Density, lbm/ft3")
axs[2].legend()

axs[3].scatter(prs_ray, py_cmix, label="python")
axs[3].set_ylabel("Speed of Sound, ft/s")
axs[3].set_xlabel("Pressure, psig")
axs[3].legend()

fig.suptitle(f"{py_mix}")

plt.show()



import matplotlib.pyplot as plt
import numpy as np

import woffl.flow.outflow as of
from woffl.flow import jetflow as jf
from woffl.flow import jetplot as jplt
from woffl.flow import singlephase as sp
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater
from woffl.pvt.resmix import ResMix

# data from MPU E-41 Well Test on 11/27/2023
# only works if the command python -m tests.jpump_test is used

pwh = 210
rho_pf = 62.4  # lbm/ft3
ppf_surf = 3168  # psi, power fluid surf pressure

# testing the jet pump code on E-41
tube = Pipe(out_dia=4.5, thick=0.5)  # E-42 tubing
case = Pipe(out_dia=6.875, thick=0.5)  # E-42 casing
ann = Annulus(inn_pipe=tube, out_pipe=case)  # define the annulus

ipr_su = InFlow(qwf=246, pwf=1049, pres=1400)  # define an ipr

e41_jp = JetPump(nozzle_no="13", area_ratio="A", ken=0.03, kth=0.3, kdi=0.4)

mpu_oil = BlackOil.schrader_oil()  # class method
mpu_wat = FormWater.schrader_wat()  # class method
mpu_gas = FormGas.schrader_gas()  # class method

form_wc = 0.894
form_gor = 600  # formation gor
form_temp = 111
prop_su = ResMix(wc=form_wc, fgor=form_gor, oil=mpu_oil, wat=mpu_wat, gas=mpu_gas)

wellprof = WellProfile.schrader()

# find the minimum psu, then find maximum, calculate pdi and compare?
# also calculate pdi_of for each case / residual?

psu_min, qoil_std, te_book = jf.psu_minimize(form_temp, e41_jp.ken, e41_jp.ate, ipr_su, prop_su)
psu_max = ipr_su.pres - 10

psu_list = np.linspace(psu_min, psu_max, 10)

pte_list = []
ptm_list = []
pdi_list = []
qoil_list = []

pni = ppf_surf + sp.diff_press_static(rho_pf, wellprof.jetpump_vd)  # static

for psu in psu_list:
    pte, ptm, pdi, qoil_std, fwat_bwpd, qnz_bwpd, mach_te, prop_tm = jf.jetpump_overall(
        psu,
        form_temp,
        pni,
        rho_pf,
        e41_jp.ken,
        e41_jp.knz,
        e41_jp.kth,
        e41_jp.kdi,
        e41_jp.ath,
        e41_jp.anz,
        tube.inn_area,
        ipr_su,
        prop_su,
    )

    pte_list.append(pte)
    ptm_list.append(ptm)
    pdi_list.append(pdi)
    qoil_list.append(qoil_std)

plt.scatter(psu_list, pdi_list, label="Discharge")
plt.scatter(psu_list, ptm_list, label="Throat Mix")
plt.scatter(psu_list, pte_list, label="Throat Entry")
plt.xlabel("Suction Pressure, psig")
plt.ylabel("Pressure, psig")
plt.title("Comparison of Jet Pump Pressures Against Suction")
plt.legend()
plt.show()



import math


class InFlow:
    def __init__(self, qwf: float, pwf: float, pres: float) -> None:
        """Initialize Well InFlow Parameters

        The well inflow class is the sandface ability to produce oil.
        Either a vogel curve or productivity index can be defined from
        the specified parameters. The oil rate is used in conjuction with
        a reservoir mixtures wc and gor to calculate other components.

        Args:
            qwf (float): Oil Rate at pwf, STB/D
            pwf (float): Flowing bottomhole pressure, psig
            pres (float): Reservoir Pressure, psig

        Returns:
            Self
        """
        if pwf > pres is True:
            raise ValueError("Flowing pressure is greater than reservoir pressure")

        self.qwf = qwf
        self.pwf = pwf
        self.pres = pres

    def __repr__(self) -> str:
        return f"Tested inflow of {self.qwf} bopd at {self.pwf} psig"

    @staticmethod
    def prod_index(qwf: float, pwf: float, pres: float) -> float:
        """Productivity Index of the Wellbore

        Calculate the straight line productivity of the wellbore.

        Args:
            qwf (float): Oil Rate at pwf, STB/D
            pwf (float): Flowing bottomhole pressure, psig
            pres (float): Reservoir Pressure, psig

        Returns:
            pidx (float): Productivity Index (STB/(day*psi))
        """
        pidx = qwf / (pres - pwf)  # stb/(day*psi)
        return pidx

    @staticmethod
    def vogel_qmax(qwf: float, pwf: float, pres: float) -> float:
        """Vogel max flow of the well

        Calculate the max flow of the well with the test data. Qmax is used
        later to calculate how much the well can flow at a new pwf.

        Args:
            qwf (float): Oil Rate at pwf, stb/d
            pwf (float): Flowing bottomhole pressure, psig
            pres (float): Reservoir Pressure, psig

        Returns:
            qmax (float): Vogel Max Theoretical Flowrate

        """
        qmax = qwf / (1 - 0.2 * (pwf / pres) - 0.8 * (pwf / pres) ** 2)
        return qmax

    def oil_flow(self, pnew: float, method: str = "vogel") -> float:
        """Oil flowrate at a new pwf, pnew

        Calculate an oil production rate at a specific bottomhole pressure.
        Could always just call it pnew, for a new bottomhold pressure?

        Args:
            pnew (float): Flowing bottomhole pressure, psig, different than self
            method (str): "vogel" or "pidx" (prod index)

        Returns:
            qnew (float): Oil rate at pnew, stb/d
        """
        if pnew < 0:
            ValueError("Flowing pressure must be greater than zero")

        if pnew > self.pres:
            ValueError("Flowing pressure must be less than reservoir pressure")

        if method == "vogel":
            qmax = self.vogel_qmax(self.qwf, self.pwf, self.pres)
            qnew = qmax * (1 - 0.2 * (pnew / self.pres) - 0.8 * (pnew / self.pres) ** 2)

        else:
            pidx = self.prod_index(self.qwf, self.pwf, self.pres)
            qnew = pidx * (self.pres - pnew)

        return qnew



"""System Operations

Code that has to do with running the entire system together. A combination of the IPR,
PVT, JetPump and Outflow. Used to create a final solution to compare.
"""

import numpy as np

from woffl.flow import jetflow as jf
from woffl.flow import jetplot as jplt
from woffl.flow import outflow as of
from woffl.flow import singlephase as sp
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.resmix import ResMix


def discharge_residual(
    psu: float,
    pwh: float,
    tsu: float,
    rho_pf: float,
    ppf_surf: float,
    jpump: JetPump,
    wellbore: Pipe,
    wellprof: WellProfile,
    ipr_su: InFlow,
    prop_su: ResMix,
) -> tuple[float, float, float, float, float]:
    """Discharge Residual

    Solve for the jet pump discharge residual, which is the difference between discharge pressure
    calculated by the jetpump and the discharge pressure from the outflow.

    Args:
        psu (float): Pressure Suction, psig
        pwh (float): Pressure Wellhead, psig
        tsu (float): Temperature Suction, deg F
        rho_pf (float): Density of the power fluid, lbm/ft3
        ppf_surf (float): Pressure Power Fluid Surface, psig
        jpump (JetPump): Jet Pump Class
        wellbore (Pipe): Pipe Class of the Wellbore
        wellprof (WellProfile): Well Profile Class
        ipr_su (InFlow): Inflow Performance Class
        prop_su (ResMix): Reservoir Mixture Conditions

    Returns:
        res_di (float): Jet Pump Discharge minus Out Flow Discharge, psid
        qoil_std (float): Oil Rate, STBOPD
        fwat_bpd (float): Formation Water Rate, BWPD
        qnz_bpd (float): Power Fluid Rate, BWPD
        mach_te (float): Throat Entry Mach, unitless
    """
    # also pump out the mach value at the throat entry?
    pni = ppf_surf + sp.diff_press_static(rho_pf, wellprof.jetpump_vd)  # static

    # jet pump section
    pte, ptm, pdi_jp, qoil_std, fwat_bwpd, qnz_bwpd, mach_te, prop_tm = jf.jetpump_overall(
        psu,
        tsu,
        pni,
        rho_pf,
        jpump.ken,
        jpump.knz,
        jpump.kth,
        jpump.kdi,
        jpump.ath,
        jpump.anz,
        wellbore.inn_area,
        ipr_su,
        prop_su,
    )

    # out flow section
    md_seg, prs_ray, slh_ray = of.top_down_press(pwh, tsu, qoil_std, prop_tm, wellbore, wellprof)

    pdi_of = prs_ray[-1]  # discharge pressure outflow
    res_di = pdi_jp - pdi_of  # what the jetpump puts out vs what is required
    return res_di, qoil_std, fwat_bwpd, qnz_bwpd, mach_te


def jetpump_solver(
    pwh: float,
    tsu: float,
    rho_pf: float,
    ppf_surf: float,
    jpump: JetPump,
    wellbore: Pipe,
    wellprof: WellProfile,
    ipr_su: InFlow,
    prop_su: ResMix,
) -> tuple[float, bool, float, float, float, float]:
    """JetPump Solver

    Find a solution for the jetpump system that factors in the wellhead pressure and reservoir conditions.
    The solver will move along the psu and dEte curves until a solution is found that satisfies the outflow
    tubing and pump conditions.

    Args:
        pwh (float): Pressure Wellhead, psig
        tsu (float): Temperature Suction, deg F
        rho_pf (float): Density of the power fluid, lbm/ft3
        ppf_surf (float): Pressure Power Fluid Surface, psig
        jpump (JetPump): Jet Pump Class
        wellbore (Pipe): Pipe Class of the Wellbore
        wellprof (WellProfile): Well Profile Class
        ipr_su (InFlow): Inflow Performance Class
        prop_su (ResMix): Reservoir Mixture Conditions

    Returns:
        psu (float): Suction Pressure, psig
        sonic_status (boolean): Is the throat entry at sonic velocity?
        qoil_std (float): Oil Rate, STBOPD
        fwat_bwpd (float): Formation Water Rate, BWPD
        qnz_bwpd (float): Power Fluid Rate, BWPD
        mach_te (float): Throat Entry Mach, unitless
    """
    psu_min, qoil_std, te_book = jf.psu_minimize(tsu=tsu, ken=jpump.ken, ate=jpump.ate, ipr_su=ipr_su, prop_su=prop_su)
    res_min, qoil_std, fwat_bwpd, qnz_bwpd, mach_te = discharge_residual(
        psu_min, pwh, tsu, rho_pf, ppf_surf, jpump, wellbore, wellprof, ipr_su, prop_su
    )

    # if the jetpump (available) discharge is above the outflow (required) discharge at lowest suction
    # the well will flow, but at its critical limit
    if res_min > 0:
        sonic_status = True
        return psu_min, sonic_status, qoil_std, fwat_bwpd, qnz_bwpd, mach_te

    psu_max = ipr_su.pres - 10  # max suction pressure that can be used
    res_max, *etc = discharge_residual(psu_max, pwh, tsu, rho_pf, ppf_surf, jpump, wellbore, wellprof, ipr_su, prop_su)

    # if the jetpump (available) discharge is below the outflow (required) discharge at highest suction
    # the well will not flow, need to pick different parameters
    if res_max < 0:
        sonic_status = False  # add code that if the flow is no, return np.NaN
        return np.nan, sonic_status, np.nan, np.nan, np.nan, np.nan

    # start secant hunting for the answer, in between the two points
    psu_list = [psu_min, psu_max]
    res_list = [res_min, res_max]

    psu_diff = 5  # criteria for when you've converged to an answer
    n = 0  # loop counter

    while abs(psu_list[-2] - psu_list[-1]) > psu_diff:
        psu_nxt = jf.psu_secant(psu_list[0], psu_list[1], res_list[0], res_list[1])
        res_nxt, qoil_std, fwat_bwpd, qnz_bwpd, mach_te = discharge_residual(
            psu_nxt, pwh, tsu, rho_pf, ppf_surf, jpump, wellbore, wellprof, ipr_su, prop_su
        )
        psu_list.append(psu_nxt)
        res_list.append(res_nxt)
        n += 1
        if n == 10:
            raise ValueError("Suction Pressure for Overall System did not converge")
    return psu_list[-1], False, qoil_std, fwat_bwpd, qnz_bwpd, mach_te



"""Two Phase Flow Equations

Equations that are predominately used in two phase flows. These are typically
the raw functions that are built together to assess two phase flow parameters.
"""

import math

import numpy as np

from woffl.flow import singlephase as sp


def velocities(qoil: float, qwat: float, qgas: float, area: float) -> tuple[float, float, float]:
    """Velocities for Two Phase Flow Calcs

    Superficial Liquid, Superficial Gas and Mixture Velocities.

    Args:
        qoil (float): Insitu Oil Volumetric Flow, ft3/s
        qwat (float): Insitu Water Volumetric Flow, ft3/s
        qgas (float): Insitu Gas Volumetric Flow, ft3/s
        area (float): Cross Sectional Area, ft2

    Returns:
        vsl (float): Superficial Liquid Velocity, ft/s
        vsg (float): Superficial Gas Velocity, ft/s
        vmix (float): Mixture Velocity, ft/s
    """
    vsl = sp.velocity((qoil + qwat), area)
    vsg = sp.velocity(qgas, area)
    vmix = vsl + vsg
    return vsl, vsg, vmix


# NFr, NLv, NGv, Nd, NRe
# HLn, HLs_0, HLs_a
def sigmoid(x: float, L: float, x0: float, k: float, b: float) -> float:
    """Sigmoid Function for Curve Fitting

    Args:
        x (float): Input data
        L (float): Scales Output Range from [0, 1] to [0, L]
        x0 (float): Middle point of Sigmoid on x-axis
        k (float): Scales the input, remains in (-inf, inf)
        b (float): Output bias, changing range from [0, L] to [b, L + b]

    Returns:
        y (float): Sigmoid Function Output
    """
    y = L / (1 + np.exp(-k * (x - x0))) + b
    return y


def ros_nlv(vsl: float, rho_liq: float, sig_liq: float) -> float:
    """Ros Dimensionless Liquid Velocity

    Args:
        vsl (float): Superficial Liquid Velocity, ft/s
        rho_liq (float): Liquid Density, lbm/ft3
        sig_liq (float): Liquid Surface Tension, lbf/ft

    Return:
        NLv (float): Ros Dimensionless Liquid Velocity
    """
    g = 32.174  # ft/s2
    sig_liq = sig_liq * g  # lbm/s2, unit conversion
    ros_nlv = vsl * (rho_liq / (g * sig_liq)) ** (1 / 4)
    return ros_nlv


def ros_ngv(vsg: float, rho_liq: float, sig_liq: float) -> float:
    """Ros Dimensionless Gas Velocity

    Args:
        vsg (float): Superficial Gas Velocity, ft/s
        rho_liq (float): Liquid Density, lbm/ft3
        sig_liq (float): Liquid Surface Tension, lbf/ft

    Return:
        NGv (float): Ros Dimensionless Liquid Velocity
    """
    g = 32.174  # ft/s2
    sig_liq = sig_liq * g  # lbm/s2, unit conversion
    ros_ngv = vsg * (rho_liq / (g * sig_liq)) ** (1 / 4)
    return ros_ngv


def ros_nd(dhyd: float, rho_liq: float, sig_liq: float) -> float:
    """Ros Dimensionless Pipe Diameter

    Args:
        dhyd (float): Hydraulic Diameter, inches
        rho_liq (float): Liquid Density, lbm/ft3
        sig_liq (float): Liquid Surface Tension, lbf/ft

    Return:
        ros_nd (float): Ros Dimensionless Pipe Diameter
    """
    g = 32.174  # ft/s2
    dhyd = dhyd / 12  # ft
    sig_liq = sig_liq * g  # lbm/s2, unit conversion
    ros_nd = dhyd * (rho_liq * g / sig_liq) ** (1 / 2)
    return ros_nd


def ros_lp(ros_nd: float) -> tuple[float, float]:
    """Ros L1 and L2 Bubble to Slug Transistion Parameter

    Ros did not provide an equation. Only a figure was provided to look at.
    The sigmoid function was used for a curve fit of the Ros figure.

    Args:
        ros_nd (float): Ros Pipe Diameter Number

    Return:
        ros_l1 (float): Ros L1 number
        ros_l2 (float): Ros L2 number

    References:
        - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 93
    """
    ros_l1 = sigmoid(ros_nd, 1.02, 43.39, -0.139, 0.99)
    ros_l2 = sigmoid(ros_nd, 0.698, 39.287, 0.093, 0.46)
    return ros_l1, ros_l2


def ros_flow_pattern(ros_ngv: float, ros_nlv: float, ros_nd: float) -> str:
    """Ros Vertical Flow Pattern

    Args:
        ros_ngv (float): Ros Gas Velocity Number, unitless
        ros_nlv (float): Ros Liquid Velocity Number, unitless
        ros_nd (float): Ros Pipe Diameter Number, unitless

    Returns:
        vpat (str): Vertical Flow Pattern

    References:
        - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 92
        - Two Phase Flow in Pipes (1988) Beggs and Brill, Page 3-31"""
    l1, l2 = ros_lp(ros_nd)
    bound_bs = l1 + l2 * ros_nlv  # bubble slug boundary
    bound_st = 50 + 36 * ros_nlv  # slug transistion boundary
    bound_tm = 75 + 84 * ros_nlv**0.75  # transition mist boundary
    # print(f"L1: {l1}, L2 {l2}, Lb: {bound_bs}, Ls: {bound_st}, Lm: {bound_tm}")

    if ros_ngv <= bound_bs:
        vpat = "bubble"
    elif (bound_bs < ros_ngv) and (ros_ngv <= bound_st):
        vpat = "slug"
    elif (bound_st < ros_ngv) and (ros_ngv <= bound_tm):
        vpat = "transistion"
    elif bound_tm < ros_ngv:
        vpat = "mist"
    else:
        # I should throw an error here
        vpat = "unknown"

    return vpat


def froude(vmix: float, dhyd: float) -> float:
    """Froude Number

    Args:
        vmix (float): Homogenous Mixture Velocity, ft/s
        dhyd (float): Hydraulic Diameter, inches

    Returns:
        froude (float): Froude Number, unitless
    """
    g = 32.174  # ft/s2
    d = dhyd / 12  # feet
    froude = vmix**2 / (g * d)
    return froude


def beggs_flow_pattern(nslh: float, froude: float) -> tuple[str, float]:
    """Beggs and Brill Horizontal Flow Pattern

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless

    Return:
        hpat (str): Horizontal Flow Pattern
        tran (float): Transitional Flow Interpolating Parameter

    References:
        - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 52
    """
    l1 = 316 * nslh**0.302
    l2 = 0.000925 * nslh**-2.468
    l3 = 0.1 * nslh**-1.468
    l4 = 0.5 * nslh**-6.738

    tran = 1  # zero for all flow patterns except for transitional

    if (nslh < 0.01) and (froude < l1):
        hpat = "segregated"
    elif (nslh >= 0.01) and (froude < l2):
        hpat = "segregated"
    elif (nslh >= 0.01) and (l2 <= froude) and (froude <= l3):
        hpat = "transition"  # can I pack the A value in here?
        tran = (l3 - froude) / (l3 - l2)
    elif (0.01 <= nslh) and (nslh < 0.4) and (l3 <= froude) and (froude <= l1):
        hpat = "intermittent"
    elif (nslh >= 0.4) and (l3 <= froude) and (froude <= l4):
        hpat = "intermittent"
    elif (nslh < 0.4) and (froude >= l1):
        hpat = "distributed"
    elif (nslh >= 0.4) and (froude > l4):
        hpat = "distributed"
    else:
        # need to thrown an error here
        hpat = "unknown"

    return hpat, tran


def beggs_holdup_base(nslh: float, froude: float, a: float, b: float, c: float) -> float:
    """Beggs and Brill Liquid Holdup Horizontal Piping

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless
        a (float): Empirical Horizontal Flow Coefficient
        b (float): Empirical Horizontal Flow Coefficient
        c (float): Empirical Horizontal Flow Coefficient

    Return:
        hlh (float): Slip Horizontal Liquid Holdup

    References:
        - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 58 (error in eqn)
    """
    hlh = a * nslh**b / froude**c
    return hlh


def beggs_holdup_horz(nslh: float, froude: float) -> tuple[float, float, float]:
    """Beggs and Brill Liquid Holdup Horizontal Piping

    Horizontal Holdup with no incline. Calculates all three to be used in interpolating
    if the flow pattern is transistional.

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless

    Return:
        hlh_seg (float): Segregated Slip Horizontal Liquid Holdup
        hlh_int (float): Intermittent Slip Horizontal Liquid Holdup
        hlh_dis (float): Distributed Slip Horizontal Liquid Holdup
    """
    hlh_seg = beggs_holdup_base(nslh, froude, 0.98, 0.4846, 0.0868)
    hlh_int = beggs_holdup_base(nslh, froude, 0.845, 0.5351, 0.0173)
    hlh_dis = beggs_holdup_base(nslh, froude, 1.065, 0.5824, 0.0609)
    return hlh_seg, hlh_int, hlh_dis


def beggs_cf_base(nslh: float, froude: float, ros_nlv: float, e: float, f: float, g: float, h: float) -> float:
    """Beggs and Brill C Factor Base

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless
        ros_nlv (float): Ros Dimensionless Liquid Velocity
        e (float): Beggs Empirical Inclined Coefficient
        f (float): Beggs Empirical Inclined Coefficient
        g (float): Beggs Empirical Inclined Coefficient
        h (float): Beggs Empirical Inclined Coefficient

    Return:
        c (float): Beggs C Factor
    """
    c = (1 - nslh) * math.log(e * (nslh**f) * (ros_nlv**g) * (froude**h))
    return c


def beggs_cf(nslh: float, froude: float, ros_nlv: float) -> tuple[float, float, float, float]:
    """Beggs and Brill C Factor Equation

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless
        ros_nlv (float): Ros Dimensionless Liquid Velocity

    Return:
        c_seg (float): Segregated Beggs C Factor
        c_int (float): Intermittent
        c_dis (float): Distributed
        c_down (float): Dowhill
    """
    c_seg = beggs_cf_base(nslh, froude, ros_nlv, 0.0110, -3.7680, 3.5390, -1.6140)
    c_int = beggs_cf_base(nslh, froude, ros_nlv, 2.9600, 0.3050, -0.4473, 0.0978)
    c_dis = 0
    c_down = beggs_cf_base(nslh, froude, ros_nlv, 4.7, -0.3692, 0.1244, -0.5056)
    return c_seg, c_int, c_dis, c_down


def beggs_phi(c: float, incline: float) -> float:
    """Beggs and Brill Phi

    The phi term is actually psi, but the nomenclature was changed since psi looks
    too much like pounds per square inch when reading it quickly.

    Args:
        c (float): Beggs C Factor
        incline (float): Angle from Horizontal, degrees

    Return:
        phi (float): Liquid Holdup Correction for Inclined Piping
    """
    inc_rad = math.radians(incline)
    phi = 1 + c * (math.sin(1.8 * inc_rad) - 0.333 * (math.sin(1.8 * inc_rad)) ** 3)
    return phi


# inc looks like int...
def beggs_holdup_inc(nslh: float, froude: float, ros_nlv: float, incline: float, hpat: str, tparm: float) -> float:
    """Beggs and Brill Incline Holdup

    Args:
        nslh (float): No Slip Liquid Holdup, unitless
        froude (float): Froude Number, unitless
        ros_nlv (float): Ros Dimensionless Liquid Velocity
        incline (float): Pipe Angle from Horizontal, degrees
        hpat (str): Horizontal Flow Pattern
        tparm (float): Transistional Flow Interpolating Parameter

    Return:
        ilh (float): Inclined Liquid Holdup
    """
    hlh_seg, hlh_int, hlh_dis = beggs_holdup_horz(nslh, froude)
    c_list = list(beggs_cf(nslh, froude, ros_nlv))
    phi_seg, phi_int, phi_dis, phi_down = [beggs_phi(c, incline) for c in c_list]

    # inclined liquid holdup, calculated for each flow regime
    ilh_up = {
        "segregated": hlh_seg * phi_seg,
        "intermittent": hlh_int * phi_int,
        "distributed": hlh_dis * phi_dis,
        "transition": tparm * hlh_seg * phi_seg + (1 - tparm) * hlh_int * phi_int,
    }

    # declined liquid holdup
    ilh_down = {
        "segregated": hlh_seg * phi_down,
        "intermittent": hlh_int * phi_down,
        "distributed": hlh_dis * phi_down,
        "transition": tparm * hlh_seg * phi_down + (1 - tparm) * hlh_int * phi_down,
    }

    if incline < 0:
        ilh = ilh_down[hpat]
    else:
        ilh = ilh_up[hpat]

    return ilh


def payne_correction(ilh: float, incline: float) -> float:
    """Payne Inclined Correction Factor

    Applies the Payne et. al. (1979) correction factor.

    Args:
        ilh (float): Inclined Liquid Holdup
        incline(float): Pipe Angle from Horizontal, degrees

    Returns:
        clh (float): Corrected Liquid Holdup
    """
    if incline > 0:
        clh = 0.924 * ilh
    elif incline < 0:
        clh = 0.685 * ilh
    else:
        clh = ilh
    return clh


def beggs_yf(nslh: float, ilh: float) -> float:
    """Beggs and Brill Y Factor

    Ratio of no slip liquid holdup to slip liquid holdup squared.
    Page 69 from Al-Safran Book.

    Args:
        nslh (float): No Slip Liquid Holdup
        ilh (float): Slip Liquid Holdup

    Returns:
        y (float): Beggs and Brill y factor
    """
    y = nslh / ilh**2
    return y


def beggs_sf(y: float) -> float:
    """Beggs and Brill S Factor

    Used for properly scalling friction factor for Beggs and Brill piping.
    Page 69 from Al-Safran Book.

    Args:
        y (float): Beggs and Brill y factor

    Returns:
        s (float): Beggs and Brill s factor
    """
    if (1 < y) and (y < 1.2):
        s = math.log(2.2 * y - 1.2)
    else:
        ln_y = math.log(y)
        s = ln_y / (-0.0523 + 3.182 * ln_y - 0.8725 * ln_y**2 + 0.01853 * ln_y**4)
    return s


def beggs_ff(fns: float, s: float) -> float:
    """Beggs and Brill Friction Factor

    Slip friction factor for Beggs and Brill piping.
    Page 69 from Al-Safran Book.

    Args:
        fns: No Slip Friction Factor, Darcy
        s (float): Beggs and Brill s factor

    Returns:
        fs (float): Slip Friction Factor, Beggs and Brill
    """
    fs = fns * math.exp(s)
    return fs


def beggs_ek(pwb: float, rho_ns: float, vmix: float, vsg: float) -> float:
    """Beggs Dimensionless Kinetic Energy

    Page 69 from Al-Safran Book, Equation 4.42

    Args:
        pwb (float): Pressure Well Bore, psig
        rho_ns (float): No Slip Mixture Density, lbm/ft3
        vmix (float): No Slip Mixture Velocity, ft/s
        vsg (float): Superficial Gas Velocity, ft/s

    Returns:
        ek (float): Beggs Dimensionless Kinetic Energy
    """
    p_abs = pwb + 14.7  # psia
    p_base = p_abs * 144 * 32.174  # lbm/(s2*ft)
    ek = vmix * vsg * rho_ns / p_base
    ek = min(ek, 0.9)  # ensure ek doesn't get bigger than 0.9
    return ek


def density_slip(rho_liq: float, rho_gas: float, slh: float) -> float:
    """Slip Mixture Density

    Slip Mixture Density, this could be placed under ResMix with the no slip
    liquid holdup as an input. It was decided against to reduce recalculating
    the density of the liquid and the density of the gas.

    Args:
        rho_liq (float): Liquid Density, lbm/ft3
        rho_gas (float): Gas Density, lbm/ft3
        slh (float): Slip Liquid Holdup, unitless

    Returns:
        rho_slip (float): Slip Mixture Density, lbm/ft3
    """
    rho_slip = rho_liq * slh + rho_gas * (1 - slh)
    return rho_slip


def beggs_press_static(rho_slip: float, height: float) -> float:
    """Beggs Static Differential Pressure

    Similiar to single phase but the density is the slip mixture.
    Positive height is up, negative height is down.

    Args:
        rho_slip (float): Slip Mixture Density, lbm/ft3
        height (float): Fluid Vertical Height, feet

    Returns:
        dp_stat (float): Static Differential Pressure, psi
    """
    dp_stat = sp.diff_press_static(rho_slip, height)
    return dp_stat


def beggs_press_friction(fb: float, rho_mix: float, vmix: float, dhyd: float, length: float) -> float:
    """Beggs Frictional Differential Pressure

    Similiar to single phase but uses beggs friction factor, and no-slip mixture
    as well as no slip density. (Which is kind of weird if we have Holdup...)

    Args:
        fb (float): Beggs friction factor of the pipe, unitless
        rho_mix (float): No Slip Mixture Density, lbm/ft3 (why use no slip...?)
        vmix (float): No Slip Mixture Velocity, ft/s
        dhyd (float): Hydraulic Diameter, inches
        length (float): Length of Pipe Segment, feet

    Returns:
        dp_fric (float): Frictional Differential Pressure, psi
    """
    dp_fric = sp.diff_press_friction(fb, rho_mix, vmix, dhyd, length)
    return dp_fric



import math

import numpy as np

import woffl.flow.singlephase as sp
from woffl.geometry.pipe import Annulus
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.formwat import FormWater


class AnnFlow:
    def __init__(
        self,
        ann_rate: float,
        surf_press: float,
        surf_temp: float,
        prop_an: FormWater,
        ann_dim: Annulus,
        wellprof: WellProfile,
    ) -> None:
        """Annular Flow

        Calculations that constrain the downward flow through the annulus.

        Args:
            ann_rate (float): Water Lift Rate, BWPD
            surf_press (float): Water Pressure at Wellhead, PSIG
            surf_temp (float): Water Temperature at Wellhead, deg F
            fluid (FormWater): Fluid to pull properties from
            ann_dim (Annulus): Annular dimensions that is flowed inside
            wellprofile (WellProfile): survey dimensions and location of jet pump

        Returns:
            Self
        """
        self.ann_rate = ann_rate
        self.surf_press = surf_press
        self.surf_temp = surf_temp
        self.prop = prop_an
        self.ann_dim = ann_dim
        self.wellprof = wellprof

        def __repr__(self):
            return f"{self.ann_rate} BWPD flowing inside a {self.ann_dim.out_pipe.inn_dia} annulus"



# import sys

import numpy as np

from woffl.pvt import BlackOil, FormGas, FormWater, ResMix

# print(sys.path)
# sys.path.append("c:\\Users\\ka9612\\OneDrive - Hilcorp\\vs_code\\hilcorpak")

mpu_oil = BlackOil.schrader()  # class method
mpu_wat = FormWater.schrader()  # class method
mpu_gas = FormGas.schrader()  # class method

e42 = ResMix(0.1, 300, mpu_oil, mpu_wat, mpu_gas)

temp = 80
press = 700

e42.condition(press, temp)

oil_comp = mpu_oil.condition(700, 80).compress()
wat_comp = mpu_wat.compress()
gas_comp = mpu_gas.condition(700, 80).compress()

print("Compressibility")
print(f"Oil: {oil_comp}, Water: {wat_comp}, Gas: {gas_comp} psi-1")

print("Bulk Modulus")
print(f"Oil: {1/oil_comp}, Water: {1/wat_comp}, Gas: {1/gas_comp} psi")

print("Speed of Sound in Mixture")
print(f"cmix: {e42.cmix()} ft/s")

print("Oil, Water, Gas Volm Fractions")
print(e42.volm_fract())



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from woffl.pvt.formgas import FormGas

# only works if the command python -m tests.fgas_test is used
dirname = os.getcwd()
filename = os.path.join(dirname, r"data\methane_hysys_peng_rob.xlsx")  # , r"data\methane_hysys_peng_rob.xlsx")

hys_df = pd.read_excel(filename, header=1)

prs_ray = hys_df["pressure"]
temp = 80
hy_rho = hys_df["density"]
hy_z_fact = hys_df["z_factor"]
hy_visc = hys_df["viscosity"]

py_rho = []
py_z_fact = []
py_visc = []
py_comp = []

py_meth = FormGas.methane_gas()

for prs in prs_ray:
    py_meth = py_meth.condition(prs, temp)
    py_rho.append(py_meth.density)
    py_z_fact.append(py_meth.zfactor())
    py_visc.append(py_meth.viscosity())
    py_comp.append(py_meth.compress())

fig, axs = plt.subplots(4, sharex=True)

axs[0].scatter(prs_ray, hy_rho, label="hysys")
axs[0].scatter(prs_ray, py_rho, label="python")
axs[0].set_ylabel("Density, lbm/ft3")
axs[0].legend()

axs[1].scatter(prs_ray, hy_visc, label="hysys")
axs[1].scatter(prs_ray, py_visc, label="python")
axs[1].set_ylabel("Viscosity, cP")
axs[1].legend()

axs[2].scatter(prs_ray, hy_z_fact, label="hysys")
axs[2].scatter(prs_ray, py_z_fact, label="python")
axs[2].set_ylabel("Z-Factor, Unitless")
axs[2].legend()

# properties of petroleum fluids, 2nd Edition, McCain, Pag 174
axs[3].scatter(1000 - 14.7, 0.001120, label="McCain, 68 deg F")
axs[3].scatter(prs_ray, py_comp, label="python")
axs[3].set_ylabel("Gas Compressibility, psi^-1")
axs[3].set_xlabel("Pressure, psig")
axs[3].legend()

fig.suptitle(f"Pure Methane Properties at {temp} deg F")
plt.show()

# blasingame 1988 PDF
blasing_zf = 0.9117
blasing_ppr = 2.958
blasing_tpr = 1.867

calc_zf = FormGas._zfactor_dak(blasing_ppr, blasing_tpr)
grad_zf = FormGas._zfactor_grad_school(blasing_ppr, blasing_tpr)
print(f"Z-Factor BG: {round(blasing_zf, 4)} DAK: {round(calc_zf, 4)} GS: {round(grad_zf, 4)}")



# import matplotlib.pyplot as plt
# import numpy as np

from woffl.geometry.wellprofile import WellProfile

# only works if the command python -m tests.wprof_test is used

sch_profile = WellProfile.schrader()
sch_profile.plot_raw()
sch_profile.plot_filter()

kup_profile = WellProfile.kuparuk()
kup_profile.plot_raw()
kup_profile.plot_filter()



class FormWater:
    def __init__(self, wat_sg: float) -> None:
        """Initialize a water stream

        Args:
            wat_sg (float): Water Specific Gravity, 0.5 to 1.5

        Returns:
            self
        """

        if (0.5 < wat_sg < 1.5) is False:
            raise ValueError(f"Water SG {wat_sg} Outside Range")

        self.wat_sg = wat_sg

    def __repr__(self) -> str:
        return f"Water {self.wat_sg} Specific Gravity"

    @classmethod
    def schrader(cls):
        """Schrader Bluff Generic Formation Water

        Args:
            wat_sg (float): 1.02"""
        return cls(wat_sg=1.02)

    @classmethod
    def kuparuk(cls):
        """Kuparuk Generic Formation Water

        Args:
            wat_sg (float): 1.02"""
        return cls(wat_sg=1.02)

    def condition(self, press: float, temp: float):
        """Set condition of evaluation

        Args:
            press (float): Pressure of the water, psig
            temp (float): Temperature of the water, deg F

        Returns:
            self
        """
        self.press = press  # psig
        self.temp = temp  # deg fahr

        self.pabs = self.press + 14.7  # convert psig to psia
        self.tabs = self.temp + 459.67  # convert fahr to rankine
        return self

    @property
    def density(self) -> float:
        """Water Density

        Calculate water density

        Args:
            None

        Returns:
            dwat (float): water density, lbm/ft3
        """

        # leave it simple now, asume no compressibility
        dwat = self.wat_sg * 62.4
        dwat = round(dwat, 3)

        self.dwat = dwat
        return dwat

    def viscosity(self) -> float:
        """Water Viscosity

        Calculate water viscosity

        Args:
            None

        Returns:
            uwat (float): water viscosity, cP
        """

        # come back later and rewrite
        uwat = 0.75  # leave as 0.75 cP for now

        self.uwat = uwat
        return uwat

    def compress(self) -> float:
        """Water Isothermal Compressibility

        Calculate isothermal water compressibility.
        Inverse of the bulk modulus of elasticity.
        Used water compressibility at 62.4 lbm/ft3 since we don't change density.

        Args:
            None

        Returns:
            wo (float): water compressibility, psi**-1

        References:
            https://roymech.org/Related/Fluids/Fluids_Water_Props.html
        """
        cw_si = 0.0004543  # 1/MPa
        cw = cw_si / 145.038  # 1/psi
        return cw

    def tension(self) -> float:
        """Water Surface Tension

        Calculate the water surface tension

        Args:
            None

        Returns:
            sigw (float): Water Surface Tension, lbf/ft
        """
        sigw_cgs = 72.8  # dyne /cm
        sigw = sigw_cgs * 0.0000685  # lbf/ft
        return sigw

    def gas_solubility(self) -> float:
        """Gas Solubility in the Water

        WOFFL assumes the formation water has no gas solubility as a
        simplifying assumption. This code is here to support other calcs.

        Args:
            None

        Returns:
            rsw (float): Gas Solubility in the water, scf/stb
        """
        return self.solubility_hawkins(self.pabs, self.temp)

    @staticmethod
    def solubility_hawkins(pabs: float, teval: float) -> float:
        """Craft and Hawkins Gas Solubility in Water

        Follows the Craft and Hawkins (1951) gas solubility method. Does
        the water need to have a bubblepoint pressure to limit this?

        Args:
            pabs (float): Eval Absolute Pressure, psia
            teval (float): Evaluated Temperature, deg F

        Returns:
            rsw (float): Gas Solubility in the water, scf/stb
        """
        a = 2.12 + 3.45e-3 * teval - 3.59e-5 * teval**2
        b = 0.0107 - 5.26e-5 * teval + 1.48e-7 * teval**2
        c = -8.75e-7 + 3.9e-9 * teval - 1.02e-11 * teval**2
        rsw = a + b * pabs + c * pabs**2
        return rsw



import matplotlib.pyplot as plt
import numpy as np

import woffl.flow.outflow as of
from woffl.geometry.pipe import Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt import BlackOil, FormGas, FormWater, ResMix

# only works if the command python -m tests.outflow_test is used
# mirror the hysys stuff
md_list = np.linspace(0, 6000, 100)
vd_list = np.linspace(0, 4000, 100)

# well with 600 fgor, 90% wc, 100 bopd
mpu_oil = BlackOil.schrader()
mpu_wat = FormWater.schrader()
mpu_gas = FormGas.schrader()
form_gor = 600  # scf/stb
form_wc = 0.9
qoil_std = 100  # stbopd

test_prop = ResMix(form_wc, form_gor, mpu_oil, mpu_wat, mpu_gas)
wellprof = WellProfile(md_list, vd_list, 6000)
tubing = Pipe(out_dia=4.5, thick=0.5)

ptop = 350  # psig
ttop = 100  # deg f

md_seg, prs_ray, slh_ray = of.top_down_press(ptop, ttop, qoil_std, test_prop, tubing, wellprof)

slh_ray = np.append(slh_ray, np.nan)  # add a nan to make same length for graphing
fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(md_seg, prs_ray, linestyle="--", color="b", label="Pressure")
ax1.set_ylabel("Pressure, PSIG")
ax1.set_xlabel("Measured Depth, Feet")
ax2.plot(md_seg, slh_ray, linestyle="-", color="r", label="Holdup")
ax2.set_ylabel("Slip Liquid Holdup")
lines1, labels1 = ax1.get_legend_handles_labels()
lines2, labels2 = ax2.get_legend_handles_labels()
ax2.legend(lines1 + lines2, labels1 + labels2, loc=0)
plt.show()

print(f"Bottom Pressure: {round(prs_ray[-1], 2)} psi")



import math

from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater


class ResMix:
    def __init__(self, wc: float, fgor: float, oil: BlackOil, wat: FormWater, gas: FormGas) -> None:
        """Reservoir Mixture

        Mixture of oil, water and natural gas. Define the watercut, fgor and classes
        of the BlackOil, FormWater and FormGas.

        Args:
            wc (float): Watercut of the Mixture, 0 to 1
            fgor (int): Formation GOR of the Mixture, scf/stb
            oil (BlackOil): Class with Oil_API, Gas_SG, Bubblepoint
            wat (FormWater): Class with Water SG
            gas (FormGas): Class with Gas SG

        Returns:
            Self
        """

        self.wc = wc  # specify a decimal
        self.fgor = fgor
        self.oil = oil
        self.wat = wat
        self.gas = gas

        # store standard densities so you don't have to call continually
        pstd = 0  # psig standard pressure
        tstd = 60  # deg f standard temperature
        self.rho_oil_std = oil.condition(pstd, tstd).density
        self.rho_wat_std = wat.condition(pstd, tstd).density
        self.rho_gas_std = gas.condition(pstd, tstd).density

    def __repr__(self) -> str:
        # return(f'Mixture with {self.oil.oil_api} API Oil')
        return f"Mixture at {100*self.wc}% Watercut and {self.fgor} SCF/STB FGOR"

    def condition(self, press: float, temp: float):
        """Set condition of evaluation

        Args:
            press (float): Pressure of the mixture, psig
            temp (float): Temperature of the mixture, deg F

        Returns:
            Self
        """
        # add code that prevents pressure from going negative?
        self.press = press
        self.temp = temp

        self.oil = self.oil.condition(press, temp)
        self.wat = self.wat.condition(press, temp)
        self.gas = self.gas.condition(press, temp)

        # input press in psig
        # input temp in deg F
        self._pressa = self.press + 14.7  # convert psig to psia
        self._tempr = self.temp + 459.67  # convert fahr to rankine
        return self

    def rho_comp(self) -> tuple[float, float, float]:
        """Density Components

        Return the density of the oil, water and gas from the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            rho_oil (float): density of oil, lbm/ft3
            rho_wat (float): density of water, lbm/ft3
            rho_gas (float): density of gas, lbm/ft3
        """
        rho_oil = self.oil.density
        rho_wat = self.wat.density
        rho_gas = self.gas.density
        return rho_oil, rho_wat, rho_gas

    def rho_two(self) -> tuple[float, float]:
        """Density of Two Phases, Liquid and Gas

        Computes the liquid mixture density. Used in multiphase equations for piping.

        Args:
            None

        Returns:
            rho_liq (float): Density of Liquid, lbm/ft3
            rho_gas (float): Density of Gas, lbm/ft3
        """
        rho_oil, rho_wat, rho_gas = self.rho_comp()
        yoil, ywat, ygas = self.volm_fract()  # volume fractions
        rho_liq = self._homogenous_liquid(yoil, ywat, rho_oil, rho_wat)
        return rho_liq, rho_gas

    def rho_mix(self) -> float:
        """Homogenous Mixture Density

        Return the homogenous density of the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            pmix (float): density of mixture, lbm/ft3
        """
        xoil, xwat, xgas = self.mass_fract()
        rho_oil, rho_wat, rho_gas = self.rho_comp()
        rho_mix = self._homogenous_density(xoil, xwat, xgas, rho_oil, rho_wat, rho_gas)
        return rho_mix

    def visc_comp(self) -> tuple[float, float, float]:
        """Viscosity Components

        Return the viscosity of the oil, water and gas from the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            uoil (float): viscosity of oil, cP
            uwat (float): viscosity of water, cP
            ugas (float): viscosity of gas, cP
        """
        uoil = self.oil.viscosity()
        uwat = self.wat.viscosity()
        ugas = self.gas.viscosity()
        return uoil, uwat, ugas

    def visc_two(self) -> tuple[float, float]:
        """Viscosity of Two Phases, Liquid and Gas

        Computes the liquid mixture viscosity. Used in multiphase equations for piping.

        Args:
            None

        Returns:
            uliq (float): Viscosity of Liquid, cP
            ugas (float): Viscosity of Gas, cP
        """
        uoil, uwat, ugas = self.visc_comp()
        yoil, ywat, ygas = self.volm_fract()  # volume fractions
        uliq = self._homogenous_liquid(yoil, ywat, uoil, uwat)
        return uliq, ugas

    def visc_mix(self) -> float:
        """Viscosity of Homogenous Mixture

        Args:
            None

        Returns:
            umix (float): Viscosity of Mixture, cP
        """
        yoil, ywat, ygas = self.volm_fract()
        uoil, uwat, ugas = self.visc_comp()
        umix = self._homogenous_mixture(yoil, ywat, ygas, uoil, uwat, ugas)
        return umix

    def tension(self) -> float:
        """Surface Tension of Liquid Phase

        Args:
            None

        Returns:
            sig_liq (float): Surface Tension Liquid, lbf/ft
        """
        sig_oil = self.oil.tension()
        sig_wat = self.wat.tension()
        yoil, ywat, ygas = self.volm_fract()  # volume fractions
        sig_liq = self._homogenous_liquid(yoil, ywat, sig_oil, sig_wat)
        return sig_liq

    def comp_comp(self) -> tuple[float, float, float]:
        """Compressibility Components

        Return the compressibility of the oil, water and gas from the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            co (float): compressibility of oil, psi**-1
            cw (float): compressibility of oil, psi**-1
            cg (float): compressibility of oil, psi**-1
        """
        co = self.oil.compress()
        cw = self.wat.compress()
        cg = self.gas.compress()
        return co, cw, cg

    def mass_fract(self) -> tuple[float, float, float]:
        """Mass Fractions

        Return the mass fractions of the oil, water and gas from the mixture.
        Requires a pressure and temperature condition to previously be set.
        Uses algebraic relationships of wc, gor and density to solve for mass fractions.

        Args:
            None

        Returns:
            xoil (float): mass fraction of oil in the mixture
            xwat (float): mass fraction of water in the mixture
            xgas (float): mass fraction of gas in the mixture

        References:
            Derivations from Kaelin available on request
        """
        # use static method from below to run the calcs
        xoil, xwat, xgas = self._owg_mass_fraction(
            self.wc, self.fgor, self.oil.gas_solubility(), self.rho_oil_std, self.rho_wat_std, self.rho_gas_std
        )
        return xoil, xwat, xgas

    def volm_fract(self) -> tuple[float, float, float]:
        """Volume Fractions

        Return the volume fractions of the oil, water and gas from the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            yoil (float): volume fraction of oil in the mixture
            ywat (float): volume fraction of water in the mixture
            ygas (float): volume fraction of gas in the mixture
        """
        xoil, xwat, xgas = self.mass_fract()
        rho_oil, rho_wat, rho_gas = self.rho_comp()
        rho_mix = self._homogenous_density(xoil, xwat, xgas, rho_oil, rho_wat, rho_gas)

        yoil = xoil * rho_mix / rho_oil
        ywat = xwat * rho_mix / rho_wat
        ygas = xgas * rho_mix / rho_gas

        return yoil, ywat, ygas

    def nslh(self) -> float:
        """No Slip Liquid Holdup

        Return the no slip liquid holdup of the mixture.

        Args:
            None

        Returns:
            nslh (float): No Slip Liquid Holdup, unitless
        """
        yoil, ywat, ygas = self.volm_fract()
        nslh = self._no_slip_liquid_holdup(yoil, ywat, ygas)
        return nslh

    def cmix(self) -> float:
        """Mixture Speed of Sound

        Return the adiabatic? Speed of Sound in the mixture.
        Requires a pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            snd_mix (float): speed of sound in the mixture, ft/s

        References:
            Sound Speed in the Mixture Water-Air D.Himr (2009)
        """
        co, cw, cg = self.comp_comp()  # isothermal compressibility
        yoil, ywat, ygas = self.volm_fract()  # volume fractions
        rho_s = self.rho_mix()
        cs = self._homogenous_mixture(yoil, ywat, ygas, co, cw, cg)  # mixture compressibility
        ks = 1 / cs  # mixture bulk modulus of elasticity
        snd_mix = math.sqrt(32.174 * 144 * ks / rho_s)  # speed of sound, ft/s
        return snd_mix

    def insitu_volm_flow(self, qoil_std: float) -> tuple[float, float, float]:
        """Insitu Volumetric Flow of Components

        Calculate the insitu volumetric flow rates of the oil, water and gas in ft3/s

        Args:
            qoil_std (float): Oil Rate, BOPD

        Returns:
            qoil (float): Oil Volumetric Flow, Insitu ft3/s
            qwat (float): Water Volumetric Flow, Insitu ft3/s
            qgas (float): Gas Volumetric Flow, Insitu ft3/s
        """
        yoil, ywat, ygas = self.volm_fract()
        qoil, qwat, qgas = self._static_insitu_volm_flow(qoil_std, self.rho_oil_std, self.oil.density, yoil, ywat, ygas)
        return qoil, qwat, qgas

    @staticmethod
    def _static_insitu_volm_flow(
        qoil_std: float, rho_oil_std: float, rho_oil: float, yoil: float, ywat: float, ygas: float
    ) -> tuple[float, float, float]:
        """Insitu Volumetric Flow of Components

        Calculate the insitu volumetric flow rates of the oil, water and gas in ft3/s

        Args:
            qoil_std (float): Oil Rate, BOPD
            rho_oil_std (float): Density Oil at Std Cond, lbm/ft3
            rho_oil (float): Density Oil Insitu Cond, lbm/ft3
            yoil (float): Volm Fraction Oil Insitu Cond, ft3/ft3
            ywat (float): Volm Fraction Water Insitu Cond, ft3/ft3
            ygas (float): Volm Fraction Gas Insitu Cond, ft3/ft3

        Returns:
            qoil (float): Oil Volumetric Flow, Insitu ft3/s
            qwat (float): Water Volumetric Flow, Insitu ft3/s
            qgas (float): Gas Volumetric Flow, Insitu ft3/s
        """
        # 42 gal/bbl, 7.48052 gal/ft3, 24 hr/day, 60min/hour, 60sec/min
        qoil_cfs = qoil_std * 42 / (24 * 60 * 60 * 7.48052)  # ft3/s at standard conditions
        moil = qoil_cfs * rho_oil_std  # mass flow of oil
        qoil = moil / rho_oil  # actual flow, ft3/s

        qtot = qoil / yoil  # oil flow divided by oil total fraction
        qwat = ywat * qtot
        qgas = ygas * qtot
        return qoil, qwat, qgas

    def insitu_mass_flow(self, qoil_std: float) -> tuple[float, float, float]:
        """Insitu Mass Flow of Components

        Calculate the insitu mass flow rates of the oil, water and gas in lbm/s
        The fractions change with pressure and temperature from the gas solubility in the oil.
        As a result the mass fractions cannot be assumed to be constant in the process.

        Args:
            qoil_std (float): Oil Rate, BOPD

        Returns:
            moil (float): Oil Mass Flow, lbm/s
            mwat (float): Water Mass Flow, lbm/s
            mgas (float): Gas Mass Flow, lbm/s
        """
        xoil, xwat, xgas = self.mass_fract()
        moil, mwat, mgas = self._static_insitu_mass_flow(qoil_std, self.rho_oil_std, xoil, xwat, xgas)
        return moil, mwat, mgas

    @staticmethod
    def _static_insitu_mass_flow(
        qoil_std: float, rho_oil_std: float, xoil: float, xwat: float, xgas: float
    ) -> tuple[float, float, float]:
        """Insitu Mass Flow of Components

        Calculate the insitu mass flow rates of the oil, water and gas in lbm/s
        The fractions change with pressure and temperature from the gas solubility in the oil.
        As a result the mass fractions cannot be assumed to be constant in the process.

        Args:
            qoil_std (float): Oil Rate, BOPD
            rho_oil_std (float): Density Oil at Std Cond, lbm/ft3
            xoil (float): Mass Fraction Oil Insitu Cond, lbm/lbm
            xwat (float): Mass Fraction Water Insitu Cond, lbm/lbm
            xgas (float): Mass Fraction Gas Insitu Cond, lbm/lbm

        Returns:
            moil (float): Oil Mass Flow, lbm/s
            mwat (float): Water Mass Flow, lbm/s
            mgas (float): Gas Mass Flow, lbm/s
        """
        # 42 gal/bbl, 7.48052 gal/ft3, 24 hr/day, 60min/hour, 60sec/min
        qoil_cfs = qoil_std * 42 / (24 * 60 * 60 * 7.48052)  # ft3/s at standard conditions
        moil = qoil_cfs * rho_oil_std  # mass flow of oil
        mtot = moil / xoil
        mwat = xwat * mtot
        mgas = xgas * mtot
        return moil, mwat, mgas

    @staticmethod
    def _owg_mass_fraction(
        wc: float, fgor: float, rs: float, rho_oil_std: float, rho_wat_std: float, rho_gas_std: float
    ) -> tuple[float, float, float]:
        """Oil Water and Gas Mass Fractions

        Return the mass fractions of the oil, water and gas from the mixture.
        Uses algebraic relationships of wc, gor and density to solve for mass fractions.
        Densities are evaluated at standard conditions

        Args:
            wc (float): Watercut of the Mixture, 0 to 1
            fgor (float): Formation GOR of the Mixture, scf/stb
            rs (float): Gas Solubility of the Oil, scf/stb
            rho_oil_std (float): Oil Density Standard Conditions, lbm/ft3
            rho_wat_std (float): Water Density Standard Conditions, lbm/ft3
            rho_gas_std (float): Gas Density Standard Conditions, lbm/ft3

        Returns:
            xoil (float): mass fraction of oil in the mixture
            xwat (float): mass fraction of water in the mixture
            xgas (float): mass fraction of gas in the mixture

        References:
            Derivations from Kaelin available on request
        """
        # mass based gas solubility
        mrs = (7.48 / 42) * rs * rho_gas_std / rho_oil_std
        # mass formation gas oil ratio, convert from scf/bbl to scf/cf
        mfgor = (7.48 / 42) * fgor * rho_gas_std / rho_oil_std
        # mass watercut
        mwc = rho_wat_std * wc / (rho_wat_std * wc + rho_oil_std * (1 - wc))
        # mass formation gas to liquid ratio
        mfglr = (7.48 / 42) * fgor * rho_gas_std * (1 - wc) / (wc * rho_wat_std + (1 - wc) * rho_oil_std)
        # mass fraction of gas
        xgas = mfglr / (1 + mfglr)
        # mass fraction of oil
        xoil = (1 - mwc) / (1 + mfgor * (1 - mwc))
        # mass fraction of water
        xwat = 1 - xoil - xgas
        # correct for the gas that is inside the oil
        xrs = xoil * mrs
        # calculate new mass fraction of free gas
        xgas = xgas - xrs
        # make sure mass fraction of gas is always zero or above
        xgas = max(xgas, 0)
        xoil = 1 - xwat - xgas
        # mass fractions
        return xoil, xwat, xgas

    @staticmethod
    def _homogenous_liquid(yoil: float, ywat: float, prop_oil: float, prop_wat: float) -> float:
        """Mixture Property of Homogenous Liquid

        Uses common assumption of homogenous liquid for calculating properties.
        Properties could be density, viscosity or surface tension.

        Args:
            yoil (float): Volume fraction of oil in the mixture
            ywat (float): Volume fraction of water in the mixture
            prop_oil (float): Property of the Oil
            prop_wat (float): Property of the Wat

        Return:
            prop_liq (float): Property of the Liquid

        References:
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 35
        """
        fw = ywat / (yoil + ywat)  # water fraction of the liquid
        prop_liq = prop_oil * (1 - fw) + prop_wat * fw
        return prop_liq

    @staticmethod
    def _homogenous_mixture(
        yoil: float, ywat: float, ygas: float, prop_oil: float, prop_wat: float, prop_gas: float
    ) -> float:
        """Property of Homogenous Mixture

        Uses common assumption of homogenous mixture for calculating properties.
        Only used for homogenous viscosity, density uses mass fractions instead.

        Args:
            yoil (float): Volume fraction of oil in the mixture
            ywat (float): Volume fraction of water in the mixture
            ygas (float): Volume fraction of gas in the mixture
            prop_oil (float): Property of the Oil
            prop_wat (float): Property of the Water
            prop_gas (float): Property of the Gas

        Return:
            prop_mix (float): Property of the Homogenous Mixture
        """
        prop_mix = yoil * prop_oil + ywat * prop_wat + ygas * prop_gas
        return prop_mix

    @staticmethod
    def _homogenous_density(
        xoil: float, xwat: float, xgas: float, rho_oil: float, rho_wat: float, rho_gas: float
    ) -> float:
        """Density of Homogenous Mixture

        Density uses mass fractions instead of volume fractions.
        That will prevent a weird circular logic loop from occuring.

        Args:
            xoil (float): Mass fraction of oil in the mixture
            xwat (float): Mass fraction of water in the mixture
            xgas (float): Mass fraction of gas in the mixture
            rho_oil (float): Density of the Oil
            rho_wat (float): Density of the Water
            rho_gas (float): Density of the Gas

        Return:
            rho_mix (float): Density of the Homogenous Mixture
        """
        vol_mix = (xoil / rho_oil) + (xwat / rho_wat) + (xgas / rho_gas)  # mixture specific volume
        rho_mix = 1 / vol_mix
        return rho_mix

    @staticmethod
    def _no_slip_liquid_holdup(yoil: float, ywat: float, ygas: float) -> float:
        """No Slip Liquid Holdup

        Use the volume fractions of the oil, water and gas.
        Calculate the no slip liquid holdup.

        Args:
            yoil (float): Volume fraction of oil in the mixture
            ywat (float): Volume fraction of water in the mixture
            ygas (float): Volume fraction of gas in the mixture

        Return:
            nslh (float): No Slip Liquid Holdup
        """
        nslh = (yoil + ywat) / (yoil + ywat + ygas)
        return nslh



import math


class FormGas:
    # ideal gas constant
    _R = 10.73  # psia*ft^3/(lbmol*Rankine)

    def __init__(self, gas_sg) -> None:
        """Initialize a Formation Gas Stream

        Args:
            gas_sg (float): Free Gas Specific Gravity

        Returns:
            Self
        """

        if (0.5 < gas_sg < 1.2) is False:
            # do I need more here?
            raise ValueError(f"Gas SG {gas_sg} Outside Range")

        # define a lot of properties just on the gas' specific gravity
        self.gas_sg = gas_sg
        self.ppc, self.tpc = self._sutton_pseudo_crit(gas_sg)
        self.mw = 28.96443 * gas_sg  # molecular weight

    def __repr__(self):
        return f"Gas: {self.gas_sg} SG and {self.mw} Mol Weight"

    @classmethod
    def schrader(cls):
        """Schrader Bluff Generic Formation Gas

        Args:
            gas_sg (float): 0.65"""
        return cls(gas_sg=0.65)

    @classmethod
    def kuparuk(cls):
        """Kuparuk Generic Formation Gas

        Args:
            gas_sg (float): 0.65"""
        return cls(gas_sg=0.65)

    @classmethod
    def methane_gas(cls):
        """Pure Methane Formation Gas

        Args:
            gas_sg (float): 0.65"""
        return cls(gas_sg=0.55)

    # almost need seperate function to change pressure / temperature
    def condition(self, press, temp):
        """Set condition of evaluation

        Args:
            press (float): Pressure of the gas, psig
            temp (float): Temperature of the gas, deg F

        Returns:
            Self
        """
        self.press = press
        self.temp = temp

        self.pabs = press + 14.7  # convert psig to psia
        self.tabs = temp + 459.67  # convert fahr to rankine

        # not adjusted for non-hydrocarbon gas such as H2S or CO2
        self.ppr = self.pabs / self.ppc  # unitless, pressure pseudo reduced
        self.tpr = self.tabs / self.tpc  # unitless, temperature pseudo reduced
        return self

    def zfactor(self) -> float:
        """Gas Z-Factor Compressibility

        Args:
            None

        Returns:
            zfactor(float): gas zfactor, no units
        """
        zfactor = self._zfactor_grad_school(self.ppr, self.tpr)
        # zfactor = self._zfactor_dak(self.ppr, self.tpr)
        return zfactor

    @property
    def density(self) -> float:
        """Gas Density, lbm/ft3

        Return the density of gas.

        Args:
            None

        Returns:
            dgas (float): density of the gas, lbm/ft3

        References:
            Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 16
            Applied Multiphase Flow in Pipes..., Al-Safran and Brill (2017) Page 305
        """
        zval = self.zfactor()  # call method if it hasn't been already?
        dgas = self.pabs * self.mw / (zval * FormGas._R * self.tabs)
        return dgas

    def viscosity(self) -> float:
        """Gas Viscosity, cP

        Return the gas viscosity, cP

        Args:
            None

        Returns:
            ugas (float): gas viscosity, cP
        """
        ug = self._viscosity_lee(self.tabs, self.mw, self.density)
        return ug

    def compress(self) -> float:
        """Gas Compressibility Isothermal

        Calculate isothermal gas compressibility.

        Args:
            None

        Returns:
            cg (float): gas compressibility, psi**-1

        References:
            Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 16
            Applied Multiphase Flow in Pipes..., Al-Safran and Brill (2017) Page 305
        """
        temp = self.temp
        p1 = self.press
        p2 = p1 + 10  # add 100 psi to evaluate a different condition for compressibility

        z1 = self.zfactor()
        z2 = self.condition(p2, temp).zfactor()

        cg = 1 / p1 - (1 / z1) * ((z2 - z1) / (p2 - p1))
        return cg

    def hydrate_check(self) -> tuple[bool, float]:
        """DO NOT USE

        Hydrate Formation Check
        Checks to see if the specified pressure and temperature is at risk of hydrates.
        Calculate hydrate formation temperature at specified pressure.
        Requires a pressure and temperature to be specified.

        Args:
            None

        Returns:
            hydrate risk (bool): True or False
            hydrate temp (float): Return Temperature of hydrate formation at pressure

        References:
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill Page 176
        """
        # placeholder for future work
        return True, 55

    @staticmethod
    def _viscosity_lee(tempr: float, mw: float, rho_gas: float) -> float:
        """Lee et. al Gas Viscosity

        Gas viscosity following Lee 1966 et al methodology.

        Args:
            tempr (float): Absolute Temperature of Gas, deg R
            mw (float): Molecular Weight of Gas Mixture, lb/lb-mol
            rho_gas (float): In-Situ Density of Gas, lbm/ft3

        Returns:
            ug (float): Gas Viscosity, cP

        References:
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 304
        """
        # tempr = temp + 459.67
        K = ((9.4 + 0.02 * mw) * tempr**1.5) / (209 + 19 * mw + tempr)
        X = 3.5 + (986 / tempr) + 0.01 * mw
        Y = 2.4 - 0.2 * X
        ug = (10**-4) * K * math.exp(X * (rho_gas / 62.4) ** Y)
        return ug

    @staticmethod
    def _sutton_pseudo_crit(gas_sg: float) -> tuple[float, float]:
        """Sutton Correlation for Pseudo Critical Pressure and Temperature

        Args:
            gas_sg (float): Gas Specific Gravity, unitless

        Returns:
            ppc (float): Pseudo Critical Pressure, psia
            tpc (float): Pseudo Critical Temperature, deg R

        References:
            - Compressibility Factors for High MW Gases (1985) Sutton, R.P, SPE 14265
        """
        ppc = 756.8 - 131.07 * gas_sg - 3.6 * gas_sg**2  # psia
        tpc = 169.2 + 349.5 * gas_sg - 74.0 * gas_sg**2  # deg R
        return ppc, tpc

    @staticmethod
    def _zfactor_grad_school(ppr: float, tpr: float) -> float:
        """Z Factor Grad School Method

        A method that was on on graduate school excel file that I borrowed.
        Much faster than the Dranchuk Abu Kassem method, unknown accuracy.

        Args:
            ppr (float): Pseudo Reduced Pressure, unitless
            tpr (float): Pseudo Reduced Pressure, unitless

        Returns:
            zfactor (float): Compressibility of Natural Gas, unitless
        """
        m = 0.51 * (tpr) ** (-4.133)
        n = 0.038 - 0.026 * (tpr) ** (1 / 2)
        zfactor = 1 - m * ppr + n * ppr**2 + 0.0003 * ppr**3
        return zfactor

    @staticmethod
    def _zfactor_dak(ppr: float, tpr: float) -> float:
        """Z Factor Dranchuk and Abu-Kassem

        Blasingame (1988) PVT Appendix has good description.

        Args:
            ppr (float): Pseudo Reduced Pressure, unitless
            tpr (float): Pseudo Reduced Temperature, unitless

        Return:
            zfactor (float): Gas Compressibility, unitless
        """
        z_ray = [0.95, 0.91]  # start calculation with a guess
        a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11 = FormGas._dak_constants()
        c1 = FormGas._dak_c1(tpr, a1, a2, a3, a4, a5)
        c2 = FormGas._dak_c2(tpr, a6, a7, a8)
        c3 = FormGas._dak_c3(tpr, a7, a8, a9)

        # need to loop this section for z-factor convergence
        while abs(z_ray[-2] - z_ray[-1]) > 0.001:
            rho_pr = FormGas._dak_rho_pr(z_ray[-1], ppr, tpr)
            c4 = FormGas._dak_c4(rho_pr, tpr, a10, a11)
            zfun = FormGas._dak_zfun(z_ray[-1], rho_pr, c1, c2, c3, c4)
            c5 = FormGas._dak_c5(z_ray[-1], rho_pr, tpr, a10, a11)
            zder = FormGas._dak_zderv(z_ray[-1], rho_pr, c1, c2, c3, c5)
            znew = z_ray[-1] - zfun / zder  # newtons method
            z_ray.append(znew)

        return z_ray[-1]

    @staticmethod
    def _dak_constants() -> tuple[float, float, float, float, float, float, float, float, float, float, float]:
        """Dranchuk and Abu-Kassem Constants

        Returns the constants used in the Dranchuk and Abu-Kassem Z factor Correlation
        """
        a1 = 0.3265
        a2 = -1.07
        a3 = -0.5339
        a4 = 0.01569
        a5 = -0.05165
        a6 = 0.5475
        a7 = -0.7361
        a8 = 0.1844
        a9 = 0.1056
        a10 = 0.6134
        a11 = 0.7210
        return a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11

    @staticmethod
    def _dak_c1(tpr: float, a1: float, a2: float, a3: float, a4: float, a5: float) -> float:
        """Dranchuk and Abu-Kassem C1

        First Part of the Dranchuk and Abu-Kassem Equation

        Args:
            tpr (float): Pseudo Reduced Temperature, unitless

        Returns:
            c1 (float): Dranchuk and Abu-Kassem C1
        """
        c1 = a1 + a2 / tpr + a3 / tpr**3 + a4 / tpr**4 + a5 / tpr**5
        return c1

    @staticmethod
    def _dak_c2(tpr: float, a6: float, a7: float, a8: float) -> float:
        """Dranchuk and Abu-Kassem C2

        Second Part of the Dranchuk and Abu-Kassem Equation

        Args:
            tpr (float): Pseudo Reduced Temperature, unitless

        Returns:
            c2 (float): Dranchuk and Abu-Kassem C2
        """
        c2 = a6 + a7 / tpr + a8 / tpr**2
        return c2

    @staticmethod
    def _dak_c3(tpr: float, a7: float, a8: float, a9: float) -> float:
        """Dranchuk and Abu-Kassem C3

        Third Part of the Dranchuk and Abu-Kassem Equation

        Args:
            tpr (float): Pseudo Reduced Temperature, unitless

        Returns:
            c3 (float): Dranchuk and Abu-Kassem C3
        """
        c3 = a9 * (a7 / tpr + a8 / tpr**2)
        return c3

    @staticmethod
    def _dak_c4(rho_pr: float, tpr: float, a10: float, a11: float) -> float:
        """Dranchuk and Abu-Kassem C4

        Fourth Part of the Dranchuk and Abu-Kassem Equation

        Args:
            rho_pr (float): Density Pseudo Reduced, units...?
            tpr (float): Temperature Pseudo Reduced, unitless

        Returns:
            c4 (float): Dranchuk and Abu-Kassem C4
        """
        c4 = a10 * (1 + a11 * rho_pr**2) * (rho_pr**2 / tpr**3) * math.exp(-a11 * rho_pr**2)
        return c4

    @staticmethod
    def _dak_rho_pr(zf: float, ppr: float, tpr: float) -> float:
        """Pseudo Reduced Density

        Used in the Dranchuk and Abu-Kassem Calculation

        Args:
            zf (float): zfactor, unitless
            ppr (float): Pressure Pseudo Reduced, unitless
            tpr (float): Temperature Pseudo Reduced, unitless

        Returns:
            rho_pr (float): Density Pseudo Reduced, units...?
        """
        rho_pr = 0.27 * ppr / (zf * tpr)
        return rho_pr

    @staticmethod
    def _dak_zfun(zf: float, rho_pr: float, c1: float, c2: float, c3: float, c4: float) -> float:
        """Calculated Z Factor with Dranchuk and Abu-Kassem Calculation

        Used to compare to the guess value to create a residual to minimize.
        Outputs the residual of the guess vs calculated zfactor. Needed to minimize.

        Args:
            zf (float): Input zfactor, unitless
            rho_pr (float): Density Pseudo Reduced, units...?
            tpr (float): Temperature Pseudo Reduced, unitless
            c1 (float): DAK C1, unitless
            c2 (float): DAK C2, unitless
            c3 (float): DAK C3, unitless
            c4 (float): DAK C4, unitless

        Return:
            zres (float): Residual of the Input zfactor vs calculated
        """
        zres = zf - (1 + c1 * rho_pr + c2 * rho_pr**2 - c3 * rho_pr**5 + c4)
        return zres

    @staticmethod
    def _dak_c5(zf: float, rho_pr: float, tpr: float, a10: float, a11: float) -> float:
        """Dranchuk and Abu-Kassem C4

        Used to compare to the guess value to create a residual to minimize.
        Add in the zf, which is the guess value, get the residual?

        Args:
            zf (float): zfactor, unitless
            rho_pr (float): Density Pseudo Reduced, units...?
            tpr (float): Temperature Pseudo Reduced, unitless
            a10 (float): DAK a1, unitless
            a11 (float): DAK a2, unitless

        Return:
            c5 (float): DAK C5, used in derivative equation
        """
        r1 = 2 * a10 * rho_pr**2 / (tpr**3 * zf)
        r2 = 1 + a11 * rho_pr**2 - (a11 * rho_pr**2) ** 2
        c5 = r1 * r2 * math.exp(-a11 * rho_pr**2)
        return c5

    @staticmethod
    def _dak_zderv(zf: float, rho_pr: float, c1: float, c2: float, c3: float, c5: float) -> float:
        """Z Factor Derivative from Dranchuk and Abu-Kassem

        Analytical derivative of the DAK zfun residual function.
        Used in Newtons Method.

        Args:
            zf (float): zfactor
            rho_pr (float): Density Pseudo Reduced, units...?
            tpr (float): Temperature Pseudo Reduced, unitless
            c1 (float): DAK C1, unitless
            c2 (float): DAK C2, unitless
            c3 (float): DAK C3, unitless
            c5 (float): DAK C5, unitless

        Returns:
            zderv (float): Z Factor Derivative
        """
        zderv = 1 + c1 * rho_pr / zf + 2 * c2 * rho_pr**2 / zf - 5 * c3 * rho_pr**5 / zf + c5
        return zderv
